/* 
   ddr1394 - An IEEE 1394 Digital Disk Recorder for IEC 61883 streams

   Copyright (C) 2002 Dan Dennedy

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

#include <termios.h>
#include <grp.h>
#include <pwd.h>
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <sys/types.h>
#include <argp.h>
#include "system.h"
#include <sys/poll.h>
#include <signal.h>

#include <libraw1394/raw1394.h>
#include "avc.h"
#include "iec61883.h"
#include "plug.h"
#include "rom.h"
#include "filelist.h"

#define EXIT_FAILURE 1

#if ENABLE_NLS
# include <libintl.h>
# define _(Text) gettext (Text)
#else
# define textdomain(Domain)
# define _(Text) Text
#endif
#define N_(Text) Text

char *xmalloc();
char *xrealloc();
char *xstrdup();

static error_t parse_opt(int key, char *arg, struct argp_state *state);
static void show_version(FILE * stream, struct argp_state *state);

/* argp option keys */
enum { DUMMY_KEY = 129, DIRECTORY_KEY
};

/* Option flags and variables.  These are initialized in parse_opt.  */

char *desired_directory;	/* --directory=DIR */
int want_verbose;		/* --verbose */

static struct argp_option options[] = {
    {"verbose", 'v', NULL, 0,
     N_("Print more information"), 0},
    {"directory", DIRECTORY_KEY, N_("DIR"), 0,
     N_("Use directory DIR"), 0},
    {NULL, 0, NULL, 0, NULL, 0}
};

/* The argp functions examine these global variables.  */
const char *argp_program_bug_address = "<dan@dennedy.org>";
void (*argp_program_version_hook) (FILE *, struct argp_state *) =
    show_version;

static struct argp argp = {
    options, parse_opt, N_("[FILE...]"),
    N_("An IEEE 1394 Digital Disk Recorder for IEC 61883 streams"),
    NULL, NULL, NULL
};


int g_done = 0;
const char not_compatible[] = "\n"
    "This libraw1394 does not work with your version of Linux. You need a different\n"
    "version that matches your kernel (see kernel help text for the raw1394 option to\n"
    "find out which is the correct version).\n";

const char not_loaded[] = "\n"
    "This probably means that you don't have raw1394 support in the kernel or that\n"
    "you haven't loaded the raw1394 module.\n";

void signal_handler(int sig)
{
    /* replace this signal handler with the default (which aborts) */
    signal(SIGINT, SIG_IGN);
    signal(SIGHUP, SIG_IGN);
    g_done = 1;
}


int main(int argc, char **argv)
{
    raw1394handle_t handle;
    struct pollfd raw1394_poll;

    /* setup signal handlers so an interruption results in cleanup */
    signal(SIGINT, signal_handler);
    signal(SIGHUP, signal_handler);

    textdomain(PACKAGE);
    argp_parse(&argp, argc, argv, 0, NULL, NULL);

    handle = raw1394_new_handle();

    if (!handle) {
	if (!errno) {
	    printf(not_compatible);
	} else {
	    perror("couldn't get handle");
	    printf(not_loaded);
	}
	exit(EXIT_FAILURE);
    }

    if (raw1394_set_port(handle, 0) < 0) {
	perror("couldn't set port");
	exit(EXIT_FAILURE);
    }

    ConfigROM rom(handle);
    Connection con(handle);
    FileList l(desired_directory);
    IEC61883Stream *ts;
    ts = new MPEG1394Stream(handle, 0, l);
    AvcTarget target(handle, ts);

#if 1
    raw1394_poll.fd = raw1394_get_fd(handle);
    raw1394_poll.events = POLLIN;
    raw1394_poll.revents = 0;

    printf("Starting; press Ctrl+C to quit...\n");
    while (!g_done) {
	if (poll(&raw1394_poll, 1, 100) > 0) {
	    if ((raw1394_poll.revents & POLLIN)
		|| (raw1394_poll.revents & POLLPRI))
		raw1394_loop_iterate(handle);
	}
    }
#else
    while (!g_done) {
	raw1394_loop_iterate(handle);
    }
#endif

    ts->stop();
    delete ts;

    printf("Quitting.\n");
    return 0;
}

/* Parse a single option.  */
static error_t parse_opt(int key, char *arg, struct argp_state *state)
{
    switch (key) {
    case ARGP_KEY_INIT:
	/* Set up default values.  */
	desired_directory = "./";
	want_verbose = 0;
	break;

    case 'v':			/* --verbose */
	want_verbose = 1;
	break;

    case DIRECTORY_KEY:	/* --directory */
	desired_directory = strdup(optarg);
	break;

    case ARGP_KEY_ARG:		/* [FILE]... */
	/* TODO: Do something with ARG, or remove this case and make
	   main give argp_parse a non-NULL fifth argument.  */
	break;

    default:
	return ARGP_ERR_UNKNOWN;
    }
    return 0;
}

/* Show the version number and copyright information.  */
static void show_version(FILE * stream, struct argp_state *state)
{
    (void) state;
    /* Print in small parts whose localizations can hopefully be copied
       from other programs.  */
    fputs(PACKAGE " " VERSION "\n", stream);
    fprintf(stream, _("Written by %s.\n\n"), "Dan Dennedy");
    fprintf(stream, _("Copyright (C) %s %s\n"), "2002", "Dan Dennedy");
    fputs(_("\
This program is free software; you may redistribute it under the terms of\n\
the GNU General Public License.  This program has absolutely no warranty.\n"), stream);
}
