/*
 *  filelist.cc - Multifile I/O class
 *
 *  Copyright (C) 2002 Dan Dennedy <dan@dennedy.org>
 *   
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <iostream>
using std::cerr;
using std::cout;
using std::endl;

#include <assert.h>
#include <stdio.h>
#include <unistd.h>
#include <dirent.h>
#include <math.h>
#include <time.h>
#include <string.h>
#include "filelist.h"

static int filterDir(const struct dirent *item)
{
    return (item->d_type == DT_REG);
}


FileList::FileList(const char *dirName):
m_dirName(dirName), m_pos(0), m_file(NULL), m_fileLength(0)
{
    chdir(m_dirName.c_str());
    rescan();
}

void FileList::rescan(void)
{
    struct dirent **eps;
    int n;

    m_list.erase(m_list.begin(), m_list.end());

    n = scandir("./", &eps, filterDir, alphasort);
    if (n >= 0) {
	int cnt;
	for (cnt = 0; cnt < n; ++cnt)
	    m_list.push_back(eps[cnt]->d_name);
    } else
	perror("Couldn't open the directory");

    if (m_list.size() > 0 && m_pos < m_list.size()) {
	if (m_file != NULL) {
	    fclose(m_file);
	    m_file = NULL;
	}

	m_file = fopen(m_list[m_pos].c_str(), "rb");
	fseek(m_file, 0, SEEK_END);
	m_fileLength = ftell(m_file);
	rewind(m_file);
    } else
	m_file = NULL;
}

FILE *FileList::nextFile(void)
{
    if (m_file != NULL && m_pos < m_list.size() - 1) {
	fclose(m_file);
	m_file = NULL;
	cout << "getting next file " << m_list[m_pos + 1] << endl;
	m_file = fopen(m_list[++m_pos].c_str(), "rb");
	fseek(m_file, 0, SEEK_END);
	m_fileLength = ftell(m_file);
	rewind(m_file);
    } else if (m_file != NULL) {
	fclose(m_file);
	m_file = NULL;
    }
    return m_file;
}

FILE *FileList::prevFile(void)
{
    if (m_file != NULL && m_pos > 0) {
	fclose(m_file);
	m_file = NULL;
	cout << "getting previous file " << m_list[m_pos - 1] << endl;
	m_file = fopen(m_list[--m_pos].c_str(), "rb");
	fseek(m_file, 0, SEEK_END);
	m_fileLength = ftell(m_file);
	rewind(m_file);
    } else if (m_file != NULL) {
	fclose(m_file);
	m_file = NULL;
    }
    return m_file;
}


bool FileList::seek(size_t length, float rate)
{
    assert(m_file != NULL);
#if 1
    static float residual = 0.0;

    /* playback rate */
    if (rate > 0.0) {
	residual += (rate - truncf(rate));
	rate = truncf(rate);
	if (residual >= 1.0) {
	    rate += 1.0;
	    residual = 0.0;
	}
    } else if (rate < 0.0) {
	residual += (rate - truncf(rate));
	rate = truncf(rate);
	if (residual <= -1.0) {
	    rate -= 1.0;
	    residual = 0.0;
	}
	cout << "rate = " << rate << ", residual = " << residual << endl;
    }
#endif
    off_t cur = ftell(m_file);

    if (rate >= 1.0 && cur > (m_fileLength - length)) {
	nextFile();
    } else if (rate <= -1.0 && (cur + (off_t) ((rate - 1.0) * length)) < 0) {
	prevFile();
	fseek(m_file, 0, SEEK_END);
	rate -= 1.0;
    }
    if (m_file == NULL)
	return false;
    if (rate == 0.0 && cur == 0) {
	if (fseek(m_file, length, SEEK_CUR) < 0) {
	    perror("failed to seek in file");
	    return false;
	}

    } else if (rate == 0.0 && cur != 0) {
	if (fseek(m_file, (-1 * length), SEEK_CUR) < 0) {
	    perror("failed to seek in file");
	    return false;
	}
    } else {
	if (fseek(m_file, (off_t) ((rate - 1.0) * length), SEEK_CUR) < 0) {
	    perror("failed to seek in file");
	    return false;
	}
    }
    return true;
}


bool FileList::read(unsigned char *buffer, size_t length)
{
    /* read a frame straight into the ringbuffer */
    return (fread(buffer, 1, length, m_file) == length);
}


bool FileList::write(unsigned char *buffer, size_t length)
{
    if (fwrite(buffer, 1, length, m_file) != length) {
	perror("error writing to file");
	return false;
    }
    m_fileLength += length;
    return true;
}

bool FileList::isAtBegin(void)
{
    if (m_file == NULL)
	return true;
    else
	return (m_pos == 0 && ftell(m_file) == 0);
}

bool FileList::isAtEnd(void)
{
    if (m_file == NULL)
	return true;
    else
	return (m_pos == (m_list.size() - 1) && feof(m_file));
}

bool FileList::newFile(const string extension)
{
    time_t t;
    struct tm now;
    char buf[32];
    buf[0] = '\1';

    time(&t);
    localtime_r(&t, &now);
    if (strftime(buf, 32, "%F_%T", &now) == 0 && buf[0] != '\0') {
	cerr << "error formatting timestamp for filename." << endl;
	return false;
    }

    strcat(buf, extension.c_str());
    m_file = fopen(buf, "wb");
    if (m_file == NULL) {
	perror("error creating new file");
	return false;
    }
    m_fileLength = 0;
    return true;
}

void FileList::commit(void)
{
    if (m_file != NULL) {
	fclose(m_file);
	m_file = NULL;
    }
    rescan();
}

void FileList::reset(void)
{
    if (m_list.size() > 0) {
	fclose(m_file);
	m_file = NULL;
	m_pos = 0;
	m_file = fopen(m_list[0].c_str(), "rb");
	fseek(m_file, 0, SEEK_END);
	m_fileLength = ftell(m_file);
	rewind(m_file);
    }
}

void FileList::ff(void)
{
    if (m_list.size() > 0) {
	fclose(m_file);
	m_file = NULL;
	m_pos = m_list.size() - 1;
	m_file = fopen(m_list[m_pos].c_str(), "rb");
	fseek(m_file, 0, SEEK_END);
	m_fileLength = ftell(m_file);
    }
}
