/*
 *  iec61883.cc - IEC 61883 Stream classes
 *
 *  Copyright (C) 2002 Mass Digital LLC and Dan Dennedy <dan@dennedy.org>
 *   
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <iostream>
using std::cout;
using std::cerr;
using std::endl;

#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/poll.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <pthread.h>
#include <math.h>
#include <errno.h>
#include <netinet/in.h>

#include <libavc1394/rom1394.h>
#include <libavc1394/plug1394.h>
#include "filelist.h"
#include "iec61883.h"


class IOError {
  private:
    int m_errno;
    string m_message;
  public:
    IOError(int error, const char *message):m_errno(error),
	m_message(message) {
    };
    int getErrno(void) {
	return m_errno;
    };
    string getMessage(void) {
	return m_message;
    };
};


IEC61883Stream::IEC61883Stream(FileList & list):
m_fileList(list), m_rate(1.0),
m_isPlaying(false), m_isRecording(false), m_cancel(true)
{
    pthread_mutex_init(&m_mutex, NULL);
}


IEC61883Stream::~IEC61883Stream(void)
{
    stop();
}


/** Static boostrap method for the thread itself.
*/
void *IEC61883Stream::transmitBootstrap(void *ptr)
{
    IEC61883Stream *me = static_cast < IEC61883Stream * >(ptr);
    me->transmitWrapper();
    return NULL;
}


/** Wrapper which ensures the correct running state.
*/
void IEC61883Stream::transmitWrapper()
{
    m_cancel = false;
    transmit();
    m_cancel = true;
}


/** Static boostrap method for the thread itself.
*/
void *IEC61883Stream::receiveBootstrap(void *ptr)
{
    IEC61883Stream *me = static_cast < IEC61883Stream * >(ptr);
    me->receiveWrapper();
    return NULL;
}


/** Wrapper which ensures the correct running state.
*/
void IEC61883Stream::receiveWrapper()
{
    m_cancel = false;
    receive();
    m_cancel = true;
}


void IEC61883Stream::play(float rate)
{
    m_rate = rate;
    if (m_fileList.currentFile() == NULL)
	m_fileList.rescan();

    if (m_fileList.currentFile() != NULL && m_cancel) {
	m_cancel = false;
	pthread_create(&m_thread, NULL, transmitBootstrap, this);
    }
}

void IEC61883Stream::pause(void)
{
    if (!m_cancel) {
	m_cancel = true;
	pthread_join(m_thread, NULL);
	if (!m_isRecording)
	    m_isPlaying = true;
    }
}

void IEC61883Stream::stop(void)
{
    if (!m_cancel || (m_isPlaying || m_isRecording)) {
	printf("IEC61883Stream::stop - Stopped Stream\n");
	m_cancel = true;
	pthread_join(m_thread, NULL);
	shutdown();
    }
}

void IEC61883Stream::record(void)
{
    if (m_cancel) {
	if (m_fileList.newFile("")) {
	    m_cancel = false;
	    printf("IEC61883Stream::record\n");
	    pthread_create(&m_thread, NULL, receiveBootstrap, this);
	}
    }
}


/////////////////////////////////////////////////////////////////////////////////////

MPEG1394Stream::MPEG1394Stream(raw1394handle_t handle, int port, FileList &list) :
IEC61883Stream(list),
m_channel(0),
m_HDVR(-1),
m_handle(0)
{
    m_handle = raw1394_new_handle();
    raw1394_set_port(m_handle, port);
}


MPEG1394Stream::~MPEG1394Stream(void)
{
    raw1394_destroy_handle(m_handle);
}


bool MPEG1394Stream::init(unsigned int channel, bool isPAL)
{
    try {
	return true;
    }
    catch(const char *message) {
	perror(message);
	shutdown();
	return false;
    }
    return true;
}

void MPEG1394Stream::stop(void)
{
    printf("STOPPING MPEG2TS STREAM!!!\n");
    m_cancel = true;
}

void MPEG1394Stream::shutdown(void)
{

    raw1394_iso_stop(m_handle);
    raw1394_iso_shutdown(m_handle);

}

#include <deque>

typedef unsigned char u8;
typedef unsigned int u32;
typedef unsigned long long u64;
typedef signed long long s64;

/*  8-byte CIP (Common Isochronous Packet) header */
struct CIP_header {
    u8 b[8];
};

static void fill_mpeg_cip_header(struct CIP_header *cip, u8 src_node_id,
				 u8 counter)
{
    cip->b[0] = src_node_id;
    cip->b[1] = 0x06;
    cip->b[2] = 0xC4;
    cip->b[3] = counter;

    cip->b[4] = 0xA0;
    cip->b[5] = 0x80;		// "time-shifted" flag; also try 0x00
    cip->b[6] = 0x00;
    cip->b[7] = 0x00;
}

/* MPEG-2 TS timestamp */
typedef u32 sph_t;

static inline u32 mpeg_ts_offset(sph_t t)
{
    return t & 0xfff;
}
static inline u32 mpeg_ts_count(sph_t t)
{
    return (t >> 12) & 0x1fff;
}
static inline sph_t make_sph(u32 count, u32 offset)
{
    return ((count & 0x1fff) << 12) | (offset & 0xfff);
}


/* MPEG-2 TSP packet (timestamp plus TS packet) */
struct TSP_packet {
    sph_t sph;
    unsigned char data[188];
};

// one ISO cycle
struct buf_cycle {
    struct CIP_header header;
    struct TSP_packet packet[3];	// up to three TSP packets
};

// 188-byte MPEG-2 TS packet
struct mpeg2_ts {
    u8 ts_header[4];
    u8 adapt_len;
    u8 adapt_flags;
    u8 pcr[6];
    u8 data[176];
};

u8 ts_get_pid(struct mpeg2_ts *ts)
{
    return ((ts->ts_header[1] << 8) + ts->ts_header[2]) & 0x1fff;
}

// returns the PCR clock, in units of 1 / 27MHz
u64 ts_get_pcr(struct mpeg2_ts * ts)
{
    u64 pcr;

    pcr = ts->pcr[0] << 25;
    pcr += ts->pcr[1] << 17;
    pcr += ts->pcr[2] << 9;
    pcr += ts->pcr[3] << 1;
    pcr += (ts->pcr[4] & 0x80) >> 7;
    pcr *= 300;
    pcr += (ts->pcr[4] & 0x1) << 8;
    pcr += ts->pcr[5];

    return pcr;
}

bool ts_has_pcr(struct mpeg2_ts * ts, int pid)
{
    if ((pid == -1) || (pid == ts_get_pid(ts))) {


	if ((ts->ts_header[3] & 0x20)	// adaptation field present
	    && (ts->adapt_len > 0)
	    && (ts->adapt_flags & 0x10)) {
	    return true;
	}
    }

    return false;
}





class TSBuffer {
  private:

    // PARAMETERS

    // max # of packets to look ahead for PCRs
    // reasonable values: 1000 - 10000
    static const unsigned int MAX_PCR_LOOKAHEAD = 20000;

    // # of PCRs to average over when estimating bitrate
    // reasonable values: 1-100
    static const unsigned int PCR_SMOOTH_INTERVAL = 20;

    // approximate # of ISO cycles transmission delay
    // valid range is 0-10; good values are 5-15
    static const unsigned int SYT_OFFSET = 7;

    // leave this off for now; I don't think it's needed with the new algorithm
    static const bool ENABLE_PCR_DRIFT_CORRECTION = false;

    // # of seconds between PCR bitrate drift checks
    // reasonable values: 1-5
    static const unsigned int PCR_DRIFT_INTERVAL = 1;


    // DATA MEMBERS

    // queue of TS packets waiting to be sent
    std::deque < struct mpeg2_ts *>ts_queue;

    // input file descriptor
    int ts_fd;

    // PCR state machine
    u64 last_pcr;		// last PCR seen

    u64 pcr_drift_ref;		// PCR from which to base drift calculations
    u32 pcr_drift_cycles;	// how many cycles have passed since pcr_drift_ref was set
    u32 packets_since_last_pcr;

    u64 delta_pcr_per_packet;

    // packetization state machine
    // num/denom algorithm for determining # of TS packets to send out in each ISO cycle
    u64 tsp_accum;
    u64 tsp_whole;
    u64 tsp_num;
    u64 tsp_denom;

    int selected_pid;
    FileList & fileList;
    float rate;

    // ISO continuity counter
    u32 iso_counter;

    // read one MPEG-2 TS packet from the input fd
    // and stick it on the end of ts_queue
    bool read_ts();

    // read packets into ts_queue until we find one with a PCR
    bool read_to_next_pcr();

    // refill the queue and update timestamper state
    bool refill();

  public:
    TSBuffer(FileList & fl, float r);

    bool init(int pid = 17);

    // output one ISO cycle's worth of packets
    // returns the total length of the ISO data
    u32 send_iso_cycle(struct buf_cycle *cycle, u32 iso_cycle,
		       u8 src_node_id);

    void setPID(int pid) {
	selected_pid = pid;
}};

TSBuffer::TSBuffer(FileList & fl, float r):fileList(fl),
last_pcr(0),
pcr_drift_ref(0),
pcr_drift_cycles(0),
tsp_accum(0), iso_counter(0), selected_pid(-1), rate(rate)
{
}


bool TSBuffer::init(int pid)
{
    selected_pid = pid;

    // skip ahead to the first PCR
    read_to_next_pcr();
    last_pcr = ts_get_pcr(ts_queue.back());

    // dump the useless packets that precede the first PCR
    while (!ts_queue.empty()) {
	delete ts_queue.front();
	ts_queue.pop_front();
    }

    return refill();
}


bool TSBuffer::read_ts()
{
    struct mpeg2_ts *new_ts = new struct mpeg2_ts;
    unsigned char *ts = (unsigned char *) new_ts;
    size_t acc = sizeof(struct mpeg2_ts);
    size_t ret;

    do {
	if (fileList.isAtEnd()) {
	    fprintf(stderr, "end of file reached\n");
	    return false;
	} else if (fileList.read(ts, 2) == false) {
	    perror("read_ts");
	    return false;
	}
    } while (ts[0] != 0x47);
    acc -= 2;
    ts += 2;
    do {
	if (fileList.isAtEnd()) {
	    fprintf(stderr, "end of file reached\n");
	    return false;
	} else if (fileList.read(ts, acc)) {
	    ts += acc;
	    acc = 0;
	} else {
	    perror("read_ts");
	    return false;
	}
    } while (acc > 0);

    ts_queue.push_back(new_ts);

    return true;
}

bool TSBuffer::read_to_next_pcr()
{
    do {
	if (ts_queue.size() > MAX_PCR_LOOKAHEAD) {
	    fprintf(stderr,
		    "couldn't find a PCR within %d packets; giving up\n",
		    MAX_PCR_LOOKAHEAD);
	    fprintf(stderr,
		    "(try reducing PCR_SMOOTH_INTERVAL or increase MAX_PCR_LOOKAHEAD\n");
	    return false;
	}

	if (!read_ts())
	    return false;

    } while (!ts_has_pcr(ts_queue.back(), selected_pid));

    return true;
}

bool TSBuffer::refill()
{
    for (unsigned int i = 0; i < PCR_SMOOTH_INTERVAL; i++)
	if (!read_to_next_pcr())
	    return false;

    u32 n_packets = ts_queue.size();

    u64 pcr = ts_get_pcr(ts_queue.back());

    if (pcr_drift_ref == 0) {
	// set up a PCR drift calculation
	pcr_drift_ref = pcr;
	pcr_drift_cycles = 0;
    }

    u64 delta_pcr = pcr - last_pcr;

    delta_pcr_per_packet = delta_pcr / n_packets;

    // Calculate the TSP packetization rate

    /* 
       This is an *exact* calculation! We want to send n_packets packets
       in the period of time covered by delta_pcr. The packet transmission
       rate should therefore be:

       (n_packets / delta_pcr) * (27,000,000 PCRs/sec)

       But what we really want is the packet transmission rate *per ISO cycle*,
       not per second. So, divide by 8000 ISO cycles/sec:

       (n_packets / delta_pcr) * (27,000,000 / sec) / (8000 cycles /sec)

       And then something magic happens: 27,000,000 / 8,000 = 3,375 exactly!

       So the *exact* transmission rate is:

       (n_packets / delta_pcr) * (3,375)        (per ISO cycle)

       We use a standard numerator/denominator algorithm to achieve this.

     */

    u64 num = n_packets * 3375;
    u64 denom = delta_pcr;

    tsp_whole = num / denom;
    tsp_num = num % denom;
    tsp_denom = denom;

    /* note: We don't reset tsp_accum to zero. This should improve our accuracy,
       as long as the transmission rate stays fairly constant. */

    if (1) {
	// for my SD streams: 1 90/2981 is best
	fprintf(stderr,
		"PCR (PID %d) after %d packets %llu delta = %llu    TSP wh %llu num %llu den %llu\n",
		ts_get_pid(ts_queue.back()), n_packets, pcr, delta_pcr,
		tsp_whole, tsp_num, tsp_denom);
    }

    last_pcr = pcr;
    packets_since_last_pcr = 0;

    return true;
}

u32 TSBuffer::send_iso_cycle(struct buf_cycle * cycle, u32 iso_cycle,
			     u8 src_node_id)
{
    // choose # of tsps by num/denom algorithm
    unsigned int n_tsps = tsp_whole;

    if (tsp_accum > (tsp_denom - tsp_num)) {
	n_tsps += 1;
	tsp_accum -= (tsp_denom - tsp_num);
    } else {
	tsp_accum += tsp_num;
    }

    if (pcr_drift_ref != 0) {
	pcr_drift_cycles++;
	if (pcr_drift_cycles % (PCR_DRIFT_INTERVAL * 8000) == 0) {

	    // time to check for PCR drift

	    // estimate the PCR that would correspond to *this* packet, if it had a PCR
	    u64 cur_pcr =
		last_pcr +
		((u64) packets_since_last_pcr) * delta_pcr_per_packet;


	    // compare cur_pcr to pcr_drift_ref + (1/27MHz) * seconds_elapsed_since_pcr_drift_ref

	    s64 drift =
		(cur_pcr - pcr_drift_ref) -
		(((s64) 27000000) * (pcr_drift_cycles / 8000));

	    if (0) {
		fprintf(stderr, "PCR drift: %lld after %u seconds\n",
			drift, pcr_drift_cycles / 8000);
	    }

	    if (ENABLE_PCR_DRIFT_CORRECTION) {
		if (pcr_drift_cycles % (5 * PCR_DRIFT_INTERVAL * 8000) ==
		    0) {
		    fprintf(stderr, "applying PCR correction of %lld\n",
			    -drift);

		    // apply the drift correction
		    last_pcr -= drift;
		    pcr_drift_ref = 0;
		}
	    }
	}
    }

    while (n_tsps > ts_queue.size())
	if (!read_ts())
	    return 0;

    // write the CIP header
    fill_mpeg_cip_header(&cycle->header, src_node_id, iso_counter);

    for (unsigned int i = 0; i < n_tsps; i++) {
	memcpy((char *) &cycle->packet[i].data[0],
	       (char *) ts_queue.front(), sizeof(struct mpeg2_ts));

	delete ts_queue.front();
	ts_queue.pop_front();

	// set timestamp to iso_cycle + SYT_OFFSET + 1000 offsets per TSP
	// (since at most 3 TSP per packet; this will ensure monotonically
	// increasing timestamps, spaced out semi-regularly)

	u32 cycle_count = (iso_cycle + SYT_OFFSET) % 8000;

	cycle->packet[i].sph = htonl(make_sph(cycle_count, 1000 * i));

	packets_since_last_pcr += 1;
    }

    if (ts_queue.empty())
	if (!refill())
	    return 0;

    // advance continuity counter by 8 per TSP in this cycle
    iso_counter += 8 * n_tsps;

    return sizeof(struct CIP_header) + n_tsps * sizeof(struct TSP_packet);
}

// set up packet smoothing buffer
static TSBuffer *buffer = NULL;

raw1394_iso_disposition MPEG1394Stream::
transmitHandler(raw1394handle_t handle, unsigned char *data,
		unsigned int *len, unsigned char *tag, unsigned char *sy,
		int cycle, unsigned int dropped)
{
    if (buffer != NULL) {
	*len = buffer->send_iso_cycle((struct buf_cycle *) data, cycle,
				      raw1394_get_local_id(handle));
    } else
	*len = 0;

    *tag = 1;
    *sy = 0;

    return RAW1394_ISO_OK;
}


void MPEG1394Stream::transmit(void)
{
    int n = 0, count = 0, pos = 0;

    try {
	buffer = new TSBuffer(getFileList(), m_rate);
	buffer->init();

	/*
	   buf_packets = 8000;
	   max_packet_size = 584;
	   channel = 63;
	   speed = RAW1394_ISO_SPEED_200;
	   irq_interval = -1;
	 */

	if (raw1394_iso_xmit_init(m_handle,
				  transmitHandler,
				  8000,
				  584, 63, RAW1394_ISO_SPEED_200, -1)) {
	    throw IOError(errno, "raw1394_iso_xmit_start");
	}

	if (raw1394_iso_xmit_start(m_handle, -1, -1)) {
	    throw IOError(errno, "raw1394_iso_xmit_start");
	}

	m_isPlaying = true;

	while (!m_fileList.isAtEnd() && !m_cancel) {
	    if (raw1394_loop_iterate(m_handle))
		throw IOError(errno, "mpeg transmit raw1394_loop_iterate");
	}

	shutdown();
	m_isPlaying = false;
	m_cancel = true;
    }
    catch(const char *message) {
	cerr << message << endl;
    }
    catch(IOError e) {
	cerr << e.getMessage() << ": " << strerror(e.getErrno()) << endl;
	shutdown();
    }

    if (buffer != NULL) {
	delete buffer;
	buffer = NULL;
    }

    return;
}


int MPEG1394Stream::findHDVR(void)
{
    int i;
    int device = -1;
    rom1394_directory rom_dir;

    printf("MPEG1394Stream::findHDVR\n");

    for (i = 0; i < raw1394_get_nodecount(m_handle); ++i) {
	if (rom1394_get_directory(m_handle, i, &rom_dir) < 0) {
	    fprintf(stderr,
		    "error reading config rom directory for node %d\n", i);
	    return -1;
	}

	if (rom_dir.vendor_id == 0x000050c9
	    && rom_dir.model_id == 0x00000002) {
	    printf("Found Tuner\n");
	    device = i;
	    break;
	}
    }

    return device;
}



static FileList *receiveFileList = NULL;

raw1394_iso_disposition MPEG1394Stream::
receiveHandler(raw1394handle_t handle, unsigned char *data,
	       unsigned int len, unsigned char channel, unsigned char tag,
	       unsigned char sy, unsigned int cycle, unsigned int dropped)
{
    unsigned short dbs_fn_qpc_sph;
    unsigned char fmt;

    /* check fields of CIP header for valid packet */
    dbs_fn_qpc_sph = (htonl(*(unsigned long *) (data)) >> 10) & 0x3fff;
    fmt = (htonl(*(unsigned long *) (data + 4)) >> 24) & 0x3f;

    if (len > 188 && fmt == 0x20 && dbs_fn_qpc_sph == 0x01b1) {
	/* skip over CIP header and SPH */
	data += 12;

	/* write each TSP in the iso packet minus SPH */
	for (; len > 188; len -= 192, data += 192)
	    if (!receiveFileList->write(data, 188)) {
		printf("Misformed packet received\n");
		return RAW1394_ISO_ERROR;
	    }
    }

    return RAW1394_ISO_OK;
}


void MPEG1394Stream::receive(void)
{
    struct oPCR o_pcr;

    try {

	receiveFileList = &getFileList();

	// TODO: connection management procedures elsewhere
	// TODO: how to generically select a device to work with?
	m_HDVR = findHDVR();
	if (m_HDVR >= 0) {
	    m_HDVR |= 0xffc0;

	    if (plug1394_get_oPCR0(m_handle, m_HDVR, &o_pcr) < 0)
		throw "failed to get oPCR";
	    o_pcr.n_p2p_connections++;
	    o_pcr.data_rate = DATA_RATE_200;
	    printf("Plug 0 is online: %d\n",
		    o_pcr.online);
	    printf("Receiving p2p transmission on channel %d\n",
		    o_pcr.channel);
	    printf("Broadcast: %d\n", o_pcr.bcast_connection);
	    printf("Payload: %d\n", o_pcr.payload);

	    if (plug1394_set_oPCR0(m_handle, m_HDVR, o_pcr) < 0)
		throw "failed to increment oPCR p2p connection count";

	    printf("oPCR setup complete\n");
	    sleep(1);
	}

	if (raw1394_iso_recv_init(m_handle,
				  receiveHandler, 1000, 584, 32, 250)) {
	    throw IOError(errno, "raw1394_iso_recv_init");
	}
	printf("raw1394_iso_recv_init successful\n");

	if (raw1394_iso_recv_start(m_handle, -1, -1, 0)) {
	    throw IOError(errno, "raw1394_iso_recv_start");
	}
	printf("raw1394_iso_recv_start successful\n");

	m_isRecording = true;

	while (!m_cancel) {
	    if (raw1394_loop_iterate(m_handle))
		throw IOError(errno, "mpeg receive raw1394_loop_iterate");

	    pthread_testcancel();
	}
	m_isRecording = false;

	if (m_HDVR >= 0) {
	    printf("Decrementing output connections count\n");
	    o_pcr.n_p2p_connections--;
	    if (plug1394_set_oPCR0(m_handle, m_HDVR, o_pcr) < 0)
		throw "failed to decrement oPCR p2p connection count";

	}
    }
    catch(const char *message) {
	cerr << message << endl;
    }
    catch(IOError e) {
	cerr << e.getMessage() << ": " << strerror(e.getErrno()) << endl;
	shutdown();
    }

    receiveFileList = NULL;
    return;
}
